import json
import locale
import logging
import re
from subprocess import run, PIPE


from plugin_utils import exception_logger, time_it

from ruxit.api.exceptions import ConfigException
from ruxit.api.base_plugin import BasePlugin
from ruxit.api.selectors import HostSelector
from ruxit.api.data import PluginMeasurement

logger = logging.getLogger(__name__)


class Quota(object):
    def __init__(self, path, size, usage):
        self.path = path
        self.size = size
        self.usage = usage
        self.pct_used = usage / size * 100

    def calculate_pct(self):
        self.pct_used = self.usage / self.size * 100

    def __str__(self):
        return f'Quota on "{self.path}" is at {self.pct_used:.2f}% ({self.usage}/{self.size}) '


def get_quotas():
    try:
        quotas = get_quotas_powershell()
    except Exception as e:
        logger.error(f"Error executing the Get-FsrmQuota command, will use dirquota instead, the error was: {e}")
        quotas = get_quotas_dirquota()
    return quotas


def get_quotas_powershell():
    quotas = []
    cmd_result = run(
        ["powershell", "(Get-FsrmQuotas).CimInstanceProperties", "|", "ConvertTo-Json"], stdout=PIPE, stderr=PIPE, check=True
    )

    if cmd_result.returncode == 0 and not cmd_result.stderr:
        try:
            raw_quotas = json.loads(cmd_result.stdout.decode("iso-8859-1"))
        except json.JSONDecodeError as e:
            logger.exception(f"Could not parse the result from {cmd_result}: {e}")
            raise ConfigException(f"Could not parse the result from {cmd_result}: {e}")

        paths = list(filter(lambda x: x["Name"] == "Path", raw_quotas))
        sizes = list(filter(lambda x: x["Name"] == "Size", raw_quotas))
        usages = list(filter(lambda x: x["Name"] == "Usage", raw_quotas))

        for data in zip(paths, sizes, usages):
            try:
                quotas.append(Quota(data[0]["Value"], locale.atof(str(data[1]["Value"])), locale.atof(str(data[2]["Value"]))))
            except:
                quotas.append(
                    Quota(
                        data[0]["Value"],
                        float(str(data[1]["Value"].replace(".", "", ",", "."))),
                        locale.atof(str(data[2]["Value"].replace(".", "", ",", "."))),
                    )
                )

        return quotas
    else:
        logger.error(f"Error executing the Get-FsrmQuota command {cmd_result.stderr}")


def get_quotas_dirquota():
    quotas = []
    cmd_result = run(["powershell", "dirquota", "quota", "list"], stdout=PIPE, stderr=PIPE, check=True)
    units = {"KB": 1_000, "MB": 1_000_000, "GB": 1_000_000_000, "TB": 1_000_000_000_000, "PB": 1_000_000_000_000_000}

    if cmd_result.returncode == 0 and not cmd_result.stderr:
        lines = cmd_result.stdout.decode("iso-8859-1").split("\r\n")
        quota = None

        for line in lines:
            if line.startswith("Quota Path:"):
                path = line.split("Quota Path:")[1].strip()
                quota = Quota(path, 1, 0)

            if quota is not None:
                # print(quota)

                if line.startswith("Limit:"):
                    value, unit = line.split("Limit:")[1].split()[:2]
                    try:
                        quota.size = locale.atof(value) * units[unit]
                    except:
                        # Sometimes python think locale is english but number comes out as portuguese
                        if value[-3] == ",":
                            quota.size = float(value.replace(".", "").replace(",", ".")) * units[unit]
                        else:
                            quota.size = float(value.replace(",", "")) * units[unit]

                if line.startswith("Used:"):
                    value, unit = line.split("Used:")[1].split()[:2]
                    try:
                        quota.usage = locale.atof(value) * units[unit]
                    except:
                        if value[-3] == ",":
                            quota.usage = float(value.replace(".", "").replace(",", ".")) * units[unit]
                        else:
                            quota.usage = float(value.replace(",", "")) * units[unit]

                if not line:
                    quota.calculate_pct()
                    quotas.append(quota)
                    quota = None

        return quotas
    else:
        logger.error(f"Error executing the dirquota command {cmd_result.stderr}")


class FSRMQuotasPlugin(BasePlugin):
    @exception_logger
    @time_it
    def query(self, **kwargs):
        config = kwargs["config"]

        logger.setLevel(logging.INFO)
        if "debug" in config and config["debug"].lower() == "true":
            logger.setLevel(logging.DEBUG)

        try:
            total_quotas = 0
            for quota in get_quotas():

                if "ignore" in config and config["ignore"]:
                    for ignore_regex in config["ignore"].split(","):
                        if re.findall(ignore_regex, quota.path):
                            logger.info(f'Ignoring quota "{quota.path}" because it matched the regex "{ignore_regex}"')

                if total_quotas >= 100:
                    logger.error(f'FSRM - Can only monitor up to 100 quotas, ignoring quota at "{quota.path}"')
                    continue

                logger.info(f"FSRM - Got quota {quota}")
                self.results_builder.add_absolute_result(
                    PluginMeasurement(
                        key="used_pct", value=quota.pct_used, dimensions={"Path": f"{quota.path}"}, entity_selector=HostSelector()
                    )
                )

                if "grab_raw_data" in config and config["grab_raw_data"].lower() == "true":
                    self.results_builder.add_absolute_result(
                        PluginMeasurement(
                            key="usage", value=quota.usage, dimensions={"Path": f"{quota.path}"}, entity_selector=HostSelector()
                        )
                    )

                    self.results_builder.add_absolute_result(
                        PluginMeasurement(
                            key="size", value=quota.size, dimensions={"Path": f"{quota.path}"}, entity_selector=HostSelector()
                        )
                    )

                if "threshold_pct" in config and config["threshold_pct"] and quota.pct_used > float(config["threshold_pct"]):
                    desc = f"The folder {quota.path} is at {quota.pct_used:.2f}% usage"[:100]
                    logger.info(f"FSRM - Sending alert {desc}")
                    self.results_builder.report_error_event(
                        description=desc,
                        title=f"Quotas Monitor - Folder Used % too high for {quota.path}",
                        entity_selector=HostSelector(),
                    )

                if "threshold_bytes" in config and config["threshold_bytes"] and quota.usage > float(config["threshold_bytes"]):
                    desc = f"The folder {quota.path} is at {quota.usage:.2f} bytes"[:100]
                    logger.info(f"FSRM - Sending alert {desc}")
                    self.results_builder.report_error_event(
                        description=desc,
                        title=f"Quotas Monitor - Folder Used Bytes too high for {quota.path}",
                        entity_selector=HostSelector(),
                    )

                total_quotas += 1
        except Exception as e:
            logger.exception(f"Error running query: {e}")
            self.results_builder.report_error_event(f"Error running fsrm_quotas_plugin: {e}", "FSRM Quotas - Exectuion Error")


def main():
    for quota in get_quotas():
        print(quota)


if __name__ == "__main__":
    main()
